/** investigate:
	* serialPersistentFields, GetField and PutField
	**/
package JSX;

import java.lang.reflect.*;
import java.io.*;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;

public class SerialPF {
	private static final int NOT_SERIALIZED = Modifier.STATIC|Modifier.TRANSIENT;
	private static final boolean DEBUG = false;
	//private static final boolean DEBUG = true;

	public static void main(String[] args) throws Exception {
		if (args.length==0) {
			args = new String[] {"java.math.BigInteger"};
			System.err.println("USAGE: java SerialPF className ["+ args[0] +"]");
			System.err.println();
		}
		Class clazz = Class.forName(args[0]);

		Map map = getFieldNameType(clazz);
		// System.err.println(map);
		for (Iterator i = map.entrySet().iterator(); i.hasNext();) {
			Map.Entry entry = (Map.Entry)i.next();
			String name = (String)entry.getKey();
			Class type = (Class)entry.getValue();
			System.err.println("\t"+type.getName()+" "+name);
		}
	}

/**	This is *only* for defWO.  Always returns true if default (only really
	* needed for serialPersistentFields).
	*
	* Check through all serializable fields (in map) to check they are
	*  - defined
	*  - not static
	*  - types match
	**/
	static public boolean areSerializableFieldsValidFor_defWO(Map map, Class clazz) {
		boolean soFarOK = true; // only needed for testing
		// check all
		for (Iterator i = map.entrySet().iterator(); i.hasNext();) {
			Map.Entry entry = (Map.Entry)i.next();
			String name = (String)entry.getKey();
			Class type = (Class)entry.getValue();
			if (DEBUG) System.err.println("\tChecking: "+type.getName()+" "+name);
		try { //for testing
			// get Field
			Field field = null;
			try {
				field = clazz.getDeclaredField(name);
			} catch (NoSuchFieldException e) {
				// if not defined...
				throw new InvalidClassException("unmatched serializable field(s) declared - '"+name+"' is not defined in "+clazz);
			}
			// if static (not an instance field)...
			if ( Modifier.isStatic(field.getModifiers()) ) {
				throw new InvalidClassException("unmatched serializable field(s) declared - '"+name+"' in "+clazz+" is static (necessarily declared in 'serialPersistentFields')");
			}
			// if types don't match...
			if (field.getType()!=type) {
				throw new InvalidClassException("unmatched serializable field(s) declared - '"+name+"' in "+clazz+" is of type "+field.getType().getName()+", but the serializable field was declared to be of type "+type.getName()+" (necessarily declared in 'serialPersistentFields')");
			}
		 } catch (InvalidClassException e) {	// for testing
			System.err.println(e);
			soFarOK = false;	// only needed for testing.
		 }
		} //...for
		return soFarOK;	// only needed for testing
	}

	// use serialPersistentFields, else defined fields (non-serializable)
	static public Map getFieldNameType(Class clazz) {
		try {
			Field spf = clazz.getDeclaredField("serialPersistentFields");
			spf.setAccessible(true);
			return getSerialPersistentFields(spf);	// get spf

		} catch (NoSuchFieldException e) {
			return getDefaultDefinedFields(clazz);	// get non-T instance fields
		}
	}

	static public boolean isFieldSerializable(Field f) {
		Class clazz = f.getDeclaringClass();	// getType() so named to distinguish
		if (Modifier.isStatic(f.getModifiers()))
			return false;	// this applies to both
		try {
			Field spf = clazz.getDeclaredField("serialPersistentFields");
			spf.setAccessible(true);
			Map map = getSerialPersistentFields(spf); // get spf
			// to do: can't check primitive type, because JSX doesn't record it.
			return map.containsKey(f.getName());	// probably should check type, too.

		} catch (NoSuchFieldException e) {	// test: is it a non-T instance field?
			if ( !Modifier.isTransient(f.getModifiers()) ) {
				return true;
			} else {
				return false;	// yes! looks ineff, but very clear & compiler zaps it!
			}
		}
	}


	static Map getSerialPersistentFields(Field spf) {
		HashMap map = new HashMap();
		ObjectStreamField[] osfs = null;
		try {
			osfs = (ObjectStreamField[])spf.get(null);	// null, because static
		} catch (IllegalAccessException e) {
			throw new InternalError(e.toString());
		}
		for (int i=0; i<osfs.length; i++) {
			ObjectStreamField f = osfs[i];
			map.put(f.getName(), f.getType());	// so can cut and paste with Field
		}	
		return map;
	}

	static HashMap getDefaultDefinedFields(Class clazz) {
		HashMap map = new HashMap();
		Field[] fields = clazz.getDeclaredFields();
		for (int i=0; i<fields.length; i++) {
			Field f = fields[i];
			if (!Modifier.isStatic(f.getModifiers()) &&
					!Modifier.isTransient(f.getModifiers()) ) {
				map.put(f.getName(), f.getType());
			}
		}
		return map;
	}

}
