/** (c) Brendan Macmillan 2002, licensed under GNU's GPL 2 
		(If your project is not GPL, then a fee is due - see license on website)
    Website: http://www.csse.monash.edu.au/~bren/JSX
    Website: http://freshmeat.net/projects/jsx
    List (can read without subscribing, need to join to post):
      http://groups.yahoo.com/group/JSX-ideas/messages
		Commercial licensing enquiries only:
			bren@objectlight.com */


package JSX.magic;
import java.io.*;
import java.lang.reflect.*;

public abstract class MagicClassI {	// we add setFinal
	protected static Field oscNameField;
	protected static ObjectStreamClass osc;
	public MagicClassI() {
		try {
			oscNameField = ObjectStreamClass.class.getDeclaredField(	//inherited
				"name");
			oscNameField.setAccessible(true);
		} catch (NoSuchFieldException noField) {
			throw new InternalError(noField.toString());
		}
	}


	abstract public Object newInstance(Class currentClass)
			throws InvocationTargetException,
						 NotSerializableException,
			       IllegalAccessException;
	abstract public void setObjectFieldValue(Object parent, Field f, Class type, Object value);
	abstract public void setPrimitiveFieldValues(Object parent, Field f, Object value);


/** Two points in XMLDeserialize that need Class.forName(), and need to send it
	* with this instead.
	* Efficiency: Could use same instance of osc each time - Caller shouldn't
	* rely on then being different objects.
	**/
	public ObjectStreamClass getOsc(String escapedName) throws ClassNotFoundException, IOException {
		setFinal(oscNameField, osc, escapedName);	//oscNameField is set statically
		return osc;
	}


//This method is never called with f==null (unless using PutFields)
//public, so magic can access...
	public void setFinal(Field f, Object parent, Object value) {
		if (f==null) return;	// if using PutFields
		if (!Modifier.isFinal(f.getModifiers())) {
			try {
				f.set(parent, value);
	 		} catch (IllegalAccessException e) {
				throw new InternalError("Unable to set '"+f+"'");
	 		}
		} else { // to set "final" fields:
			if (f.getType().isPrimitive())
				setPrimitiveFieldValues(parent, f, value);	// template method
			else
				setObjectFieldValue(parent, f, f.getType(), value);	// template method
		}
	}
}
